declare const defaultEqualityCheck: EqualityFn;
interface DefaultMemoizeOptions {
    equalityCheck?: EqualityFn;
    resultEqualityCheck?: EqualityFn;
    maxSize?: number;
}
declare function defaultMemoize<F extends AnyFunction>(func: F, equalityCheckOrOptions?: EqualityFn | DefaultMemoizeOptions): F & {
    clearCache: () => void;
};

type LongestTuple<T extends readonly unknown[][]> = T extends [
    infer U extends unknown[]
] ? U : T extends [infer U, ...infer R extends unknown[][]] ? MostProperties<U, LongestTuple<R>> : never;
type MostProperties<T, U> = keyof U extends keyof T ? T : U;
type ElementAt<T extends unknown[], N extends keyof any> = N extends keyof T ? T[N] : unknown;
type ElementsAt<T extends readonly unknown[][], N extends keyof any> = {
    [K in keyof T]: ElementAt<T[K], N>;
};
type Intersect<T extends readonly unknown[]> = T extends [] ? unknown : T extends [infer H, ...infer T] ? H & Intersect<T> : T[number];
type MergeTuples<T extends readonly unknown[][], L extends unknown[] = LongestTuple<T>> = {
    [K in keyof L]: Intersect<ElementsAt<T, K>>;
};
type ExtractParameters<T extends readonly AnyFunction[]> = {
    [K in keyof T]: Parameters<T[K]>;
};
type MergeParameters<T extends readonly AnyFunction[]> = '0' extends keyof T ? MergeTuples<ExtractParameters<T>> : Parameters<T[number]>;

/**
 * A standard selector function, which takes three generic type arguments:
 * @template State - The first value, often a Redux root state object
 * @template Result - The final result returned by the selector
 * @template Params - All additional arguments passed into the selector
 */
type Selector<State = any, Result = unknown, Params extends readonly any[] = any[]> = 
/**
 * A function that takes a state and returns data that is based on that state.
 *
 * @param state - The first argument, often a Redux root state object.
 * @param params - All additional arguments passed into the selector.
 * @returns A derived value from the state.
 */
(state: State, ...params: FallbackIfNever<Params, []>) => Result;
/**
 * A function that takes input selectors' return values as arguments and returns a result. Otherwise known as `resultFunc`.
 *
 * @template InputSelectors - An array of input selectors.
 * @template Result - Result returned by `resultFunc`.
 */
type Combiner<InputSelectors extends SelectorArray, Result> = 
/**
 * A function that takes input selectors' return values as arguments and returns a result. Otherwise known as `resultFunc`.
 *
 * @param resultFuncArgs - Return values of input selectors.
 * @returns The return value of {@linkcode OutputSelectorFields.resultFunc resultFunc}.
 */
(...resultFuncArgs: SelectorResultArray<InputSelectors>) => Result;
/**
 * The additional fields attached to the output selector generated by `createSelector`.
 *
 * **Note**: Although {@linkcode CreateSelectorOptions.memoize memoize}
 * and {@linkcode CreateSelectorOptions.argsMemoize argsMemoize} are included in the attached fields,
 * the fields themselves are independent of the type of
 * {@linkcode CreateSelectorOptions.memoize memoize} and {@linkcode CreateSelectorOptions.argsMemoize argsMemoize} functions.
 * Meaning this type is not going to generate additional fields based on what functions we use to memoize our selectors.
 *
 * _This type is not to be confused with {@linkcode ExtractMemoizerFields ExtractMemoizerFields}._
 *
 * @template InputSelectors - The type of the input selectors.
 * @template Result - The type of the result returned by the `resultFunc`.
 * @template MemoizeFunction - The type of the memoize function that is used to memoize the `resultFunc` inside `createSelector` (e.g., `defaultMemoize` or `weakMapMemoize`).
 * @template ArgsMemoizeFunction - The type of the optional memoize function that is used to memoize the arguments passed into the output selector generated by `createSelector` (e.g., `defaultMemoize` or `weakMapMemoize`). If none is explicitly provided, `defaultMemoize` will be used.
 */
interface OutputSelectorFields<InputSelectors extends SelectorArray = SelectorArray, Result = unknown, MemoizeFunction extends UnknownMemoizer = typeof defaultMemoize, ArgsMemoizeFunction extends UnknownMemoizer = typeof defaultMemoize> extends Required<Pick<CreateSelectorOptions<MemoizeFunction, ArgsMemoizeFunction>, 'argsMemoize' | 'memoize'>> {
    /** The final function passed to `createSelector`. Otherwise known as the `combiner`. */
    resultFunc: Combiner<InputSelectors, Result>;
    /** The memoized version of {@linkcode resultFunc resultFunc}. */
    memoizedResultFunc: Combiner<InputSelectors, Result> & ExtractMemoizerFields<MemoizeFunction>;
    /** Returns the last result calculated by the output selector. */
    lastResult: () => Result;
    /** An array of the input selectors. */
    dependencies: InputSelectors;
    /** Counts the number of times the output has been recalculated. */
    recomputations: () => number;
    /** Resets the count of `recomputations` count to 0. */
    resetRecomputations: () => 0;
}
/**
 * Represents the actual selectors generated by `createSelector`.
 *
 * @template InputSelectors - The type of the input selectors.
 * @template Result - The type of the result returned by the `resultFunc`.
 * @template MemoizeFunction - The type of the memoize function that is used to memoize the `resultFunc` inside `createSelector` (e.g., `defaultMemoize` or `weakMapMemoize`).
 * @template ArgsMemoizeFunction - The type of the optional memoize function that is used to memoize the arguments passed into the output selector generated by `createSelector` (e.g., `defaultMemoize` or `weakMapMemoize`). If none is explicitly provided, `defaultMemoize` will be used.
 */
type OutputSelector<InputSelectors extends SelectorArray = SelectorArray, Result = unknown, MemoizeFunction extends UnknownMemoizer = typeof defaultMemoize, ArgsMemoizeFunction extends UnknownMemoizer = typeof defaultMemoize> = PrepareOutputSelector<InputSelectors, Result, MemoizeFunction, ArgsMemoizeFunction> & ExtractMemoizerFields<ArgsMemoizeFunction>;
/**
 * A helper type designed to optimize TypeScript performance by composing parts of {@linkcode OutputSelector OutputSelector} in a more statically structured manner.
 *
 * This is achieved by utilizing the `extends` keyword with `interfaces`, as opposed to creating intersections with type aliases.
 * This approach offers some performance benefits:
 * - `Interfaces` create a flat object type, while intersections with type aliases recursively merge properties.
 * - Type relationships between `interfaces` are also cached, as opposed to intersection types as a whole.
 * - When checking against an intersection type, every constituent is verified before checking against the "effective" flattened type.
 *
 * This optimization focuses on resolving much of the type composition for
 * {@linkcode OutputSelector OutputSelector} using `extends` with `interfaces`,
 * rather than relying on intersections for the entire {@linkcode OutputSelector OutputSelector}.
 *
 * @template InputSelectors - The type of the input selectors.
 * @template Result - The type of the result returned by the `resultFunc`.
 * @template MemoizeFunction - The type of the memoize function that is used to memoize the `resultFunc` inside `createSelector` (e.g., `defaultMemoize` or `weakMapMemoize`).
 * @template ArgsMemoizeFunction - The type of the optional memoize function that is used to memoize the arguments passed into the output selector generated by `createSelector` (e.g., `defaultMemoize` or `weakMapMemoize`). If none is explicitly provided, `defaultMemoize` will be used.
 *
 * @see {@link https://github.com/microsoft/TypeScript/wiki/Performance#preferring-interfaces-over-intersections Reference}
 */
interface PrepareOutputSelector<InputSelectors extends SelectorArray = SelectorArray, Result = unknown, MemoizeFunction extends UnknownMemoizer = typeof defaultMemoize, ArgsMemoizeFunction extends UnknownMemoizer = typeof defaultMemoize> extends OutputSelectorFields<InputSelectors, Result, MemoizeFunction, ArgsMemoizeFunction>, Selector<GetStateFromSelectors<InputSelectors>, Result, GetParamsFromSelectors<InputSelectors>> {
}
/**
 * A selector that is assumed to have one additional argument, such as
 * the props from a React component
 */
type ParametricSelector<State, Props, Result> = Selector<State, Result, [
    Props,
    ...any
]>;
/** A generated selector that is assumed to have one additional argument */
type OutputParametricSelector<State, Props, Result> = ParametricSelector<State, Props, Result> & OutputSelectorFields<SelectorArray, Result>;
/** An array of input selectors */
type SelectorArray = ReadonlyArray<Selector>;
/** A standard function returning true if two values are considered equal */
type EqualityFn = (a: any, b: any) => boolean;
type StabilityCheckFrequency = 'always' | 'once' | 'never';
/**
 * The options object used inside `createSelector` and `createSelectorCreator`.
 *
 * @template MemoizeFunction - The type of the memoize function that is used to memoize the `resultFunc` inside `createSelector` (e.g., `defaultMemoize` or `weakMapMemoize`).
 * @template ArgsMemoizeFunction - The type of the optional memoize function that is used to memoize the arguments passed into the output selector generated by `createSelector` (e.g., `defaultMemoize` or `weakMapMemoize`). If none is explicitly provided, `defaultMemoize` will be used.
 * @template OverrideMemoizeFunction - The type of the optional `memoize` function that could be passed into the options object inside `createSelector` to override the original `memoize` function that was initially passed into `createSelectorCreator`.
 * @template OverrideArgsMemoizeFunction - The type of the optional `argsMemoize` function that could be passed into the options object inside `createSelector` to override the original `argsMemoize` function that was initially passed into `createSelectorCreator`. If none was initially provided, `defaultMemoize` will be used.
 */
interface CreateSelectorOptions<MemoizeFunction extends UnknownMemoizer, ArgsMemoizeFunction extends UnknownMemoizer, OverrideMemoizeFunction extends UnknownMemoizer = never, OverrideArgsMemoizeFunction extends UnknownMemoizer = never> {
    /**
     * Overrides the global input stability check for the selector.
     * - `once` - Run only the first time the selector is called.
     * - `always` - Run every time the selector is called.
     * - `never` - Never run the input stability check.
     *
     * @default 'once'
     *
     * @see {@link https://github.com/reduxjs/reselect#development-only-checks development-only-checks}
     * @see {@link https://github.com/reduxjs/reselect#inputstabilitycheck inputStabilityCheck}
     * @see {@link https://github.com/reduxjs/reselect#per-selector-configuration per-selector-configuration}
     */
    inputStabilityCheck?: StabilityCheckFrequency;
    /**
     * The memoize function that is used to memoize the {@linkcode OutputSelectorFields.resultFunc resultFunc}
     * inside `createSelector` (e.g., `defaultMemoize` or `weakMapMemoize`).
     *
     * When passed directly into `createSelector`, it overrides the `memoize` function initially passed into `createSelectorCreator`.
     */
    memoize: FallbackIfNever<OverrideMemoizeFunction, MemoizeFunction>;
    /**
     * The optional memoize function that is used to memoize the arguments passed into the output selector generated by `createSelector` (e.g., `defaultMemoize` or `weakMapMemoize`).
     *
     * When passed directly into `createSelector`, it overrides the `argsMemoize` function initially passed into `createSelectorCreator`. If none was initially provided, `defaultMemoize` will be used.
     *
     * @default defaultMemoize
     */
    argsMemoize?: FallbackIfNever<OverrideArgsMemoizeFunction, ArgsMemoizeFunction>;
    /**
     * Optional configuration options for the {@linkcode memoize memoize} function.
     * These options are passed to the {@linkcode memoize memoize} function as the second argument.
     */
    memoizeOptions?: OverrideMemoizeOptions<MemoizeFunction, OverrideMemoizeFunction>;
    /**
     * Optional configuration options for the {@linkcode argsMemoize argsMemoize} function.
     * These options are passed to the {@linkcode argsMemoize argsMemoize} function as the second argument.
     */
    argsMemoizeOptions?: OverrideMemoizeOptions<ArgsMemoizeFunction, OverrideArgsMemoizeFunction>;
}
/** Extracts an array of all return types from all input selectors */
type SelectorResultArray<Selectors extends SelectorArray> = ExtractReturnType<Selectors>;
/** Determines the combined single "State" type (first arg) from all input selectors */
type GetStateFromSelectors<Selectors extends SelectorArray> = MergeParameters<Selectors>[0];
/** Determines the combined  "Params" type (all remaining args) from all input selectors */
type GetParamsFromSelectors<Selectors extends SelectorArray, RemainingItems extends readonly unknown[] = Tail<MergeParameters<Selectors>>> = RemainingItems;
/** Any function with any arguments */
type AnyFunction = (...args: any[]) => any;
/** Any function with unknown arguments */
type UnknownFunction = (...args: unknown[]) => unknown;
/** Any Memoizer function. A memoizer is a function that accepts another function and returns it. */
type UnknownMemoizer<Func extends UnknownFunction = UnknownFunction> = (func: Func, ...options: any[]) => Func;
/**
 * When a generic type parameter is using its default value of `never`, fallback to a different type.
 *
 * @template T - Type to be checked.
 * @template FallbackTo - Type to fallback to if `T` resolves to `never`.
 */
type FallbackIfNever<T, FallbackTo> = IfNever<T, FallbackTo, T>;
/**
 * Derive the type of memoize options object based on whether the memoize function itself was overridden.
 *
 * _This type can be used for both `memoizeOptions` and `argsMemoizeOptions`._
 *
 * @template MemoizeFunction - The type of the `memoize` or `argsMemoize` function initially passed into `createSelectorCreator`.
 * @template OverrideMemoizeFunction - The type of the optional `memoize` or `argsMemoize` function passed directly into `createSelector` which then overrides the original `memoize` or `argsMemoize` function passed into `createSelectorCreator`.
 */
type OverrideMemoizeOptions<MemoizeFunction extends UnknownMemoizer, OverrideMemoizeFunction extends UnknownMemoizer = never> = IfNever<OverrideMemoizeFunction, MemoizeOptionsFromParameters<MemoizeFunction>, MemoizeOptionsFromParameters<OverrideMemoizeFunction>>;
/**
 * Extract the memoize options from the parameters of a memoize function.
 *
 * @template MemoizeFunction - The type of the memoize function to be checked.
 */
type MemoizeOptionsFromParameters<MemoizeFunction extends UnknownMemoizer> = DropFirstParameter<MemoizeFunction>[0] | DropFirstParameter<MemoizeFunction>;
/**
 * Extracts the additional fields that a memoize function attaches to the function it memoizes (e.g., `clearCache`).
 *
 * @template MemoizeFunction - The type of the memoize function to be checked.
 */
type ExtractMemoizerFields<MemoizeFunction extends UnknownMemoizer> = OmitIndexSignature<ReturnType<MemoizeFunction>>;
/** Extract the return type from all functions as a tuple */
type ExtractReturnType<T extends readonly AnyFunction[]> = {
    [index in keyof T]: T[index] extends T[number] ? ReturnType<T[index]> : never;
};
/** First item in an array */
type Head<T> = T extends [any, ...any[]] ? T[0] : never;
/** All other items in an array */
type Tail<A> = A extends [any, ...infer Rest] ? Rest : never;
/**
 * An if-else-like type that resolves depending on whether the given type is `never`.
 * This is mainly used to conditionally resolve the type of a `memoizeOptions` object based on whether `memoize` is provided or not.
 * @see {@link https://github.com/sindresorhus/type-fest/blob/main/source/if-never.d.ts Source}
 */
type IfNever<T, TypeIfNever, TypeIfNotNever> = [T] extends [never] ? TypeIfNever : TypeIfNotNever;
/**
 * Omit any index signatures from the given object type, leaving only explicitly defined properties.
 * This is mainly used to remove explicit `any`s from the return type of some memoizers (e.g, `microMemoize`).
 * @see {@link https://github.com/sindresorhus/type-fest/blob/main/source/omit-index-signature.d.ts Source}
 */
type OmitIndexSignature<ObjectType> = {
    [KeyType in keyof ObjectType as {} extends Record<KeyType, unknown> ? never : KeyType]: ObjectType[KeyType];
};
/**
 * The infamous "convert a union type to an intersection type" hack
 * Source: https://github.com/sindresorhus/type-fest/blob/main/source/union-to-intersection.d.ts
 * Reference: https://github.com/microsoft/TypeScript/issues/29594
 */
type UnionToIntersection<Union> = (Union extends unknown ? (distributedUnion: Union) => void : never) extends (mergedIntersection: infer Intersection) => void ? Intersection : never;
/**
 * Code to convert a union of values into a tuple.
 * Source: https://stackoverflow.com/a/55128956/62937
 */
type Push<T extends any[], V> = [...T, V];
type LastOf<T> = UnionToIntersection<T extends any ? () => T : never> extends () => infer R ? R : never;
type TuplifyUnion<T, L = LastOf<T>, N = [T] extends [never] ? true : false> = true extends N ? [] : Push<TuplifyUnion<Exclude<T, L>>, L>;
/**
 * Converts "the values of an object" into a tuple, like a type-level `Object.values()`
 * Source: https://stackoverflow.com/a/68695508/62937
 */
type ObjValueTuple<T, KS extends any[] = TuplifyUnion<keyof T>, R extends any[] = []> = KS extends [infer K, ...infer KT] ? ObjValueTuple<T, KT, [...R, T[K & keyof T]]> : R;
/** Utility type to infer the type of "all params of a function except the first", so we can determine what arguments a memoize function accepts */
type DropFirstParameter<Func extends AnyFunction> = Func extends (firstArg: any, ...restArgs: infer Rest) => any ? Rest : never;

declare function autotrackMemoize<F extends AnyFunction>(func: F): F & {
    clearCache: () => void;
};

declare function weakMapMemoize<F extends AnyFunction>(func: F): F & {
    clearCache: () => void;
};

/**
 * An instance of `createSelector`, customized with a given memoize implementation.
 *
 * @template MemoizeFunction - The type of the memoize function that is used to memoize the `resultFunc` inside `createSelector` (e.g., `defaultMemoize` or `weakMapMemoize`).
 * @template ArgsMemoizeFunction - The type of the optional memoize function that is used to memoize the arguments passed into the output selector generated by `createSelector` (e.g., `defaultMemoize` or `weakMapMemoize`). If none is explicitly provided, `defaultMemoize` will be used.
 */
interface CreateSelectorFunction<MemoizeFunction extends UnknownMemoizer, ArgsMemoizeFunction extends UnknownMemoizer = typeof defaultMemoize> {
    /**
     * Creates a memoized selector function.
     *
     * @param createSelectorArgs - An arbitrary number of input selectors as separate inline arguments and a `combiner` function.
     * @returns An output selector.
     *
     * @template InputSelectors - The type of the input selectors as an array.
     * @template Result - The return type of the `combiner` as well as the output selector.
     * @template OverrideMemoizeFunction - The type of the optional `memoize` function that could be passed into the options object to override the original `memoize` function that was initially passed into `createSelectorCreator`.
     * @template OverrideArgsMemoizeFunction - The type of the optional `argsMemoize` function that could be passed into the options object to override the original `argsMemoize` function that was initially passed into `createSelectorCreator`.
     */
    <InputSelectors extends SelectorArray, Result>(...createSelectorArgs: [
        ...inputSelectors: InputSelectors,
        combiner: Combiner<InputSelectors, Result>
    ]): OutputSelector<InputSelectors, Result, MemoizeFunction, ArgsMemoizeFunction>;
    /**
     * Creates a memoized selector function.
     *
     * @param createSelectorArgs - An arbitrary number of input selectors as separate inline arguments, a `combiner` function and an `options` object.
     * @returns An output selector.
     *
     * @template InputSelectors - The type of the input selectors as an array.
     * @template Result - The return type of the `combiner` as well as the output selector.
     * @template OverrideMemoizeFunction - The type of the optional `memoize` function that could be passed into the options object to override the original `memoize` function that was initially passed into `createSelectorCreator`.
     * @template OverrideArgsMemoizeFunction - The type of the optional `argsMemoize` function that could be passed into the options object to override the original `argsMemoize` function that was initially passed into `createSelectorCreator`.
     */
    <InputSelectors extends SelectorArray, Result, OverrideMemoizeFunction extends UnknownMemoizer = MemoizeFunction, OverrideArgsMemoizeFunction extends UnknownMemoizer = ArgsMemoizeFunction>(...createSelectorArgs: [
        ...inputSelectors: InputSelectors,
        combiner: Combiner<InputSelectors, Result>,
        createSelectorOptions: Partial<CreateSelectorOptions<MemoizeFunction, ArgsMemoizeFunction, OverrideMemoizeFunction, OverrideArgsMemoizeFunction>>
    ]): OutputSelector<InputSelectors, Result, OverrideMemoizeFunction, OverrideArgsMemoizeFunction>;
    /**
     * Creates a memoized selector function.
     *
     * @param inputSelectors - An array of input selectors.
     * @param combiner - A function that Combines the input selectors and returns an output selector. Otherwise known as the result function.
     * @param createSelectorOptions - An optional options object that allows for further customization per selector.
     * @returns An output selector.
     *
     * @template InputSelectors - The type of the input selectors array.
     * @template Result - The return type of the `combiner` as well as the output selector.
     * @template OverrideMemoizeFunction - The type of the optional `memoize` function that could be passed into the options object to override the original `memoize` function that was initially passed into `createSelectorCreator`.
     * @template OverrideArgsMemoizeFunction - The type of the optional `argsMemoize` function that could be passed into the options object to override the original `argsMemoize` function that was initially passed into `createSelectorCreator`.
     */
    <InputSelectors extends SelectorArray, Result, OverrideMemoizeFunction extends UnknownMemoizer = MemoizeFunction, OverrideArgsMemoizeFunction extends UnknownMemoizer = ArgsMemoizeFunction>(inputSelectors: [...InputSelectors], combiner: Combiner<InputSelectors, Result>, createSelectorOptions?: Partial<CreateSelectorOptions<MemoizeFunction, ArgsMemoizeFunction, OverrideMemoizeFunction, OverrideArgsMemoizeFunction>>): OutputSelector<InputSelectors, Result, OverrideMemoizeFunction, OverrideArgsMemoizeFunction>;
}
/**
 * In development mode, an extra check is conducted on your input selectors.
 * It runs your input selectors an extra time with the same arguments, and
 * warns in the console if they return a different result _(based on your `memoize` method)_.
 *
 * This function allows you to override this setting for all of your selectors.
 *
 * **Note**: This setting can still be overridden per selector inside `createSelector`'s `options` object.
 * See {@link https://github.com/reduxjs/reselect#per-selector-configuration per-selector-configuration}
 * and {@linkcode CreateSelectorOptions.inputStabilityCheck inputStabilityCheck} for more details.
 *
 * _The input stability check does not run in production builds._
 *
 * @param inputStabilityCheckFrequency - How often the `inputStabilityCheck` should run for all selectors.
 *
 * @example
 * ```ts
 * import { setInputStabilityCheckEnabled } from 'reselect'
 *
 * // Run only the first time the selector is called. (default)
 * setInputStabilityCheckEnabled('once')
 *
 * // Run every time the selector is called.
 * setInputStabilityCheckEnabled('always')
 *
 * // Never run the input stability check.
 * setInputStabilityCheckEnabled('never')
 * ```
 * @see {@link https://github.com/reduxjs/reselect#development-only-checks development-only-checks}
 * @see {@link https://github.com/reduxjs/reselect#global-configuration global-configuration}
 */
declare function setInputStabilityCheckEnabled(inputStabilityCheckFrequency: StabilityCheckFrequency): void;
/**
 * Creates a selector creator function with the specified memoization function and options for customizing memoization behavior.
 *
 * @param options - An options object containing the `memoize` function responsible for memoizing the `resultFunc` inside `createSelector` (e.g., `defaultMemoize` or `weakMapMemoize`). It also provides additional options for customizing memoization. While the `memoize` property is mandatory, the rest are optional.
 * @returns A customized `createSelector` function.
 *
 * @example
 * ```ts
 * const customCreateSelector = createSelectorCreator({
 *   memoize: customMemoize, // Function to be used to memoize `resultFunc`
 *   memoizeOptions: [memoizeOption1, memoizeOption2], // Options passed to `customMemoize` as the second argument onwards
 *   argsMemoize: customArgsMemoize, // Function to be used to memoize the selector's arguments
 *   argsMemoizeOptions: [argsMemoizeOption1, argsMemoizeOption2] // Options passed to `customArgsMemoize` as the second argument onwards
 * })
 *
 * const customSelector = customCreateSelector(
 *   [inputSelector1, inputSelector2],
 *   resultFunc // `resultFunc` will be passed as the first argument to `customMemoize`
 * )
 *
 * customSelector(
 *   ...selectorArgs // Will be memoized by `customArgsMemoize`
 * )
 * ```
 *
 * @template MemoizeFunction - The type of the memoize function that is used to memoize the `resultFunc` inside `createSelector` (e.g., `defaultMemoize` or `weakMapMemoize`).
 * @template ArgsMemoizeFunction - The type of the optional memoize function that is used to memoize the arguments passed into the output selector generated by `createSelector` (e.g., `defaultMemoize` or `weakMapMemoize`). If none is explicitly provided, `defaultMemoize` will be used.
 */
declare function createSelectorCreator<MemoizeFunction extends UnknownMemoizer, ArgsMemoizeFunction extends UnknownMemoizer = typeof defaultMemoize>(options: CreateSelectorOptions<MemoizeFunction, typeof defaultMemoize, never, ArgsMemoizeFunction>): CreateSelectorFunction<MemoizeFunction, ArgsMemoizeFunction>;
/**
 * Creates a selector creator function with the specified memoization function and options for customizing memoization behavior.
 *
 * @param memoize - The `memoize` function responsible for memoizing the `resultFunc` inside `createSelector` (e.g., `defaultMemoize` or `weakMapMemoize`).
 * @param memoizeOptionsFromArgs - Optional configuration options for the memoization function. These options are then passed to the memoize function as the second argument onwards.
 * @returns A customized `createSelector` function.
 *
 * @example
 * ```ts
 * const customCreateSelector = createSelectorCreator(customMemoize, // Function to be used to memoize `resultFunc`
 *   option1, // Will be passed as second argument to `customMemoize`
 *   option2, // Will be passed as third argument to `customMemoize`
 *   option3 // Will be passed as fourth argument to `customMemoize`
 * )
 *
 * const customSelector = customCreateSelector(
 *   [inputSelector1, inputSelector2],
 *   resultFunc // `resultFunc` will be passed as the first argument to `customMemoize`
 * )
 * ```
 *
 * @template MemoizeFunction - The type of the memoize function that is used to memoize the `resultFunc` inside `createSelector` (e.g., `defaultMemoize` or `weakMapMemoize`).
 */
declare function createSelectorCreator<MemoizeFunction extends UnknownMemoizer>(memoize: MemoizeFunction, ...memoizeOptionsFromArgs: DropFirstParameter<MemoizeFunction>): CreateSelectorFunction<MemoizeFunction>;
declare const createSelector: CreateSelectorFunction<typeof defaultMemoize, typeof defaultMemoize>;

interface SelectorsObject {
    [key: string]: AnyFunction;
}
interface StructuredSelectorCreator {
    <SelectorMap extends SelectorsObject, SelectorParams = MergeParameters<ObjValueTuple<SelectorMap>>>(selectorMap: SelectorMap, selectorCreator?: CreateSelectorFunction<any, any>): (state: Head<SelectorParams>, ...params: Tail<SelectorParams>) => {
        [Key in keyof SelectorMap]: ReturnType<SelectorMap[Key]>;
    };
    <State, Result = State>(selectors: {
        [K in keyof Result]: Selector<State, Result[K], never>;
    }, selectorCreator?: CreateSelectorFunction<any, any>): Selector<State, Result, never>;
}
declare const createStructuredSelector: StructuredSelectorCreator;

export { CreateSelectorFunction, CreateSelectorOptions, DefaultMemoizeOptions, EqualityFn, GetParamsFromSelectors, GetStateFromSelectors, OutputParametricSelector, OutputSelector, OutputSelectorFields, ParametricSelector, Selector, SelectorArray, SelectorResultArray, StructuredSelectorCreator, createSelector, createSelectorCreator, createStructuredSelector, defaultEqualityCheck, defaultMemoize, setInputStabilityCheckEnabled, autotrackMemoize as unstable_autotrackMemoize, weakMapMemoize };
